@@ -1,4 +1,4 @@
-__version__ = '1.3'
+__version__ = '1.4'
 
 class SpotifyError(Exception):
     pass
@@ -14,6 +14,8 @@ from _spotify import PlaylistContainer
 from _spotify import AlbumBrowser
 from _spotify import ArtistBrowser
 from _spotify import Image
+from _spotify import User
+from _spotify import ToplistBrowser
 
 from _spotify import api_version
 from _spotify import connect
