@@ -79,87 +79,25 @@ plc_callbacks_table_add(PlaylistContainer * plc,
     }
 }
 
-static playlistcontainer_callback *
-plc_callbacks_table_remove(PlaylistContainer * plc, PyObject *callback,
-                           PyObject *userdata)
-{
-    plc_cb_entry *e_prev = NULL, *e_curr, *entry = NULL;
-    playlistcontainer_callback *c_prev = NULL, *c_curr;
-    playlistcontainer_callback *result = NULL;
-    PyObject *code1, *code2;
-
-    /* Look for an existing entry for this playlist */
-    e_curr = playlistcontainer_callbacks_table;
-    while (e_curr) {
-        if (e_curr->playlistcontainer == plc->_playlistcontainer) {
-            entry = e_curr;
-            break;
-        }
-        e_prev = e_curr;
-        e_curr = e_curr->next;
-    }
-    /* Update callbacks entry */
-    if (!entry) {
-        return NULL;
-    }
-    else {
-        c_curr = entry->callbacks;
-        while (c_curr) {
-            /* Note: a sole Python function can be represented by several
-             * Python Function objects. However, to each function corresponds
-             * an unique Code object.
-             */
-            code1 = PyFunction_GetCode(c_curr->trampoline->callback);
-            code2 = PyFunction_GetCode(callback);
-            if (code1 == code2 && c_curr->trampoline->userdata == userdata) {
-                result = c_curr;
-                if (c_prev) {
-                    c_prev->next = c_curr->next;
-                }
-                else {
-                    entry->callbacks = c_curr->next;
-                };
-                break;
-            }
-            c_prev = c_curr;
-            c_curr = c_curr->next;
-        }
-    }
-    if (!result)
-        return NULL;
-    /* Cleanup */
-    if (!entry->callbacks) {
-        if (e_prev) {
-            e_prev->next = entry->next;
-        }
-        else {
-            playlistcontainer_callbacks_table = entry->next;
-        }
-        sp_playlistcontainer_release(entry->playlistcontainer);
-        free(entry);
-    }
-    return result;
-}
-
 static PyObject *
 PlaylistContainer_add_callback(PlaylistContainer * self,
                                PyObject *args,
                                sp_playlistcontainer_callbacks * plc_callbacks)
 {
-    PyObject *callback;
-    PyObject *manager = NULL, *userdata = NULL;
+    PyObject *callback, *userdata = NULL;
     Callback *tramp;
-    playlistcontainer_callback *head, *to_add;
+    playlistcontainer_callback *to_add;
 
-    if (!PyArg_ParseTuple(args, "O|OO", &callback, &manager, &userdata))
+    if (!PyArg_ParseTuple(args, "O|O", &callback, &userdata))
         return NULL;
     if (!userdata)
         userdata = Py_None;
-    if (!manager)
-        manager = Py_None;
-    if (!(callback = as_function(callback)))
+    if (!(PyFunction_Check(callback) || PyMethod_Check(callback))) {
+        PyErr_SetString(PyExc_TypeError,
+                    "callback argument must be of function or method type");
         return NULL;
-    tramp = create_trampoline(callback, manager, userdata);
+    }
+    tramp = create_trampoline(callback, Py_None, userdata);
     to_add = malloc(sizeof(playlistcontainer_callback));
     to_add->callback = plc_callbacks;
     to_add->trampoline = tramp;
@@ -185,8 +123,9 @@ playlistcontainer_loaded_callback(sp_playlistcontainer * playlistcontainer,
     gstate = PyGILState_Ensure();
     pc = PlaylistContainer_FromSpotify(playlistcontainer);
     res = PyObject_CallFunctionObjArgs(tramp->callback,
-                                       tramp->manager,
                                        pc, tramp->userdata, NULL);
+    if (!res)
+        PyErr_WriteUnraisable(tramp->callback);
     Py_DECREF(pc);
     Py_XDECREF(res);
     PyGILState_Release(gstate);
@@ -219,8 +158,9 @@ playlistcontainer_playlist_added_callback(sp_playlistcontainer *
     p = Playlist_FromSpotify(playlist);
     pos = Py_BuildValue("i", position);
     res = PyObject_CallFunctionObjArgs(tramp->callback,
-                                       tramp->manager,
                                        pc, p, pos, tramp->userdata, NULL);
+    if (!res)
+        PyErr_WriteUnraisable(tramp->callback);
     Py_DECREF(pc);
     Py_DECREF(p);
     Py_DECREF(pos);
@@ -259,9 +199,10 @@ playlistcontainer_playlist_moved_callback(sp_playlistcontainer *
     pos = Py_BuildValue("i", position);
     new_pos = Py_BuildValue("i", new_position);
     res = PyObject_CallFunctionObjArgs(tramp->callback,
-                                       tramp->manager,
                                        pc,
                                        p, pos, new_pos, tramp->userdata, NULL);
+    if (!res)
+        PyErr_WriteUnraisable(tramp->callback);
     Py_DECREF(pc);
     Py_DECREF(p);
     Py_DECREF(pos);
@@ -299,8 +240,9 @@ playlistcontainer_playlist_removed_callback(sp_playlistcontainer *
     p = Playlist_FromSpotify(playlist);
     pos = Py_BuildValue("i", position);
     res = PyObject_CallFunctionObjArgs(tramp->callback,
-                                       tramp->manager,
                                        pc, p, pos, tramp->userdata, NULL);
+    if (!res)
+        PyErr_WriteUnraisable(tramp->callback);
     Py_DECREF(pc);
     Py_DECREF(p);
     Py_DECREF(pos);
@@ -324,8 +266,6 @@ PlaylistContainer_add_playlist_removed_callback(PlaylistContainer * self,
 static PyObject *
 PlaylistContainer_str(PyObject *o)
 {
-    PlaylistContainer *self = (PlaylistContainer *) o;
-
     PyErr_SetString(PyExc_NotImplementedError, "");
     return NULL;
 }
@@ -338,6 +278,37 @@ PlaylistContainer_sq_length(PyObject *o)
     return sp_playlistcontainer_num_playlists(self->_playlistcontainer);
 }
 
+static PyObject *
+PlaylistContainer_add_new_playlist(PlaylistContainer *pc, PyObject *args)
+{
+    char *name = NULL;
+    int len;
+    sp_playlist *playlist;
+
+    if (!sp_playlistcontainer_is_loaded(pc->_playlistcontainer)) {
+        PyErr_SetString(SpotifyError, "PlaylistContainer not loaded");
+        return NULL;
+    }
+    if (!PyArg_ParseTuple(args, "es#", ENCODING, &name, &len))
+        return NULL;
+    if (len > 255) {
+        PyErr_SetString(PyExc_ValueError,
+                "Playlist name must be < 255 characters long");
+        goto error;
+    }
+    playlist =
+        sp_playlistcontainer_add_new_playlist(pc->_playlistcontainer, name);
+    if (!playlist) {
+        PyErr_SetString(SpotifyError, "Operation failed.");
+        goto error;
+    }
+    return Playlist_FromSpotify(playlist);
+
+    error:
+        PyMem_Free(name);
+        return NULL;
+}
+
 /// PlaylistContainer Get Item []
 PyObject *
 PlaylistContainer_sq_item(PyObject *o, Py_ssize_t index)
@@ -391,6 +362,10 @@ static PyMethodDef PlaylistContainer_methods[] = {
      (PyCFunction)PlaylistContainer_add_playlist_removed_callback,
      METH_VARARGS,
      ""},
+    {"add_new_playlist",
+     (PyCFunction)PlaylistContainer_add_new_playlist,
+     METH_VARARGS,
+     "Add a new empty playlist to the playlist container."},
     {NULL}
 };
 
