@@ -1,25 +1,11 @@
-# $Id$
-#
-# Copyright 2009 Doug Winter
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-# http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
+# encoding: utf-8
 
 import unittest
 from spotify._mockspotify import mock_track, mock_album, mock_artist, mock_session
 
 class TestTrack(unittest.TestCase):
 
-    track = mock_track("foo", 3, mock_album("bar", mock_artist("baz", 1), 0, "",
+    track = mock_track(u'æâ€êþÿ', 3, mock_album("bar", mock_artist("baz", 1), 0, "",
                                             0, 1, 1), 10, 20, 30, 40, 0, 1)
 
     def test_artists(self):
@@ -29,7 +15,7 @@ class TestTrack(unittest.TestCase):
         self.assertEqual(self.track.album().name(), "bar")
 
     def test_name(self):
-        self.assertEqual(self.track.name(), "foo")
+        self.assertEqual(self.track.name(), u'æâ€êþÿ')
 
     def test_duration(self):
         self.assertEqual(self.track.duration(), 10)
